/* NVClock 0.8 - Linux overclocker for NVIDIA cards
 *
 * Copyright(C) 2001-2007 Roderick Colenbrander, DOS version by RayeR, 2009
 *
 * site: http://nvclock.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef __DJGPP__

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <dos.h>
#include <pc.h>
#include <dpmi.h>
#include "nvreg.h"
#include "backend.h"

//***************** PCI bus constants and macros ****************************
#define PCI_CAR32	 0x0CF8        // PCI Config Address Register I/O address (32 bit)
#define PCI_CDR32	 0x0CFC        // PCI Config Data Register I/O address (32 bit)
#define PCI_MAX_BUSES	 256	       // muze byt nejvyse 256 PCI sbernic (obvykle 1-3, zacinaji od 0)
#define PCI_MAX_DEVICES  32	       // na 1 PCI sbernici muze byt nejvyse 32 zarizeni
#define PCI_MAX_FUNCS	 8	       // v jednom PCI zarizeni muze byt nejvyse 8 funkcnich bloku
#define PCI_MAX_REGS	 64	       // 1 funkcni blok muze obsahovat nejvyse 64 32-bit registru (256B)

// 32-bit CAR: bit31=1, bit30:24=reserved, bit23:16=bus number [0-255], bit15:11=device number [0-31], bit10:8=function number [0-7], bit7:2=32-bit register [0-63], bit1:0=0
#define PCI_CA32(busn,devn,funcn,regn) ((0x80000000UL)|(((uint32_t)((busn)&0xFF))<<16)|(((uint32_t)((devn)&0x1F))<<11)|(((uint32_t)((funcn)&0x07))<<8)|(((uint32_t)((regn)&0x3F))<<2))

//***************** PCI device header structure *****************************
typedef struct _pci_config_header
{
  uint16_t VendorID;
  uint16_t DeviceID;
  uint16_t Command;
  uint16_t Status;
  uint8_t  RevisionID;
  uint8_t  ProgIf;
  uint8_t  SubClass;
  uint8_t  BaseClass;
  uint8_t  CacheLineSize;
  uint8_t  LatencyTimer;
  uint8_t  HeaderType;
  uint8_t  BIST;
  uint32_t BaseAddresses[6];
  uint32_t CardBusCISPtr;
  uint16_t SubsystemVendorID;
  uint16_t SubsystemID;
  uint32_t ROMBaseAddress;
  uint32_t Reserved2[2];
  uint8_t  InterruptLine;
  uint8_t  InterruptPin;
  uint8_t  MinimumGrant;
  uint8_t  MaximumLatency;
} __attribute__((packed)) PCI_CONFIG_HEADER;

NVClock nvclock;
NVCard *nv_card = NULL;

//***************** read 32-bit PCI registr from given PCI address **********
uint32_t pci_read32(uint32_t busn, uint32_t devn, uint32_t funcn, uint32_t regn32) // regn32 [0-63]
{
  uint32_t pci_car32;
  uint32_t data;
  disable();			       // disable interrupts
  pci_car32=inportl(PCI_CAR32);        // store current CAR32
  outportl(PCI_CAR32,PCI_CA32(busn,devn,funcn,regn32)); // set required CAR32
  data=inportl(PCI_CDR32);	       // read 32bit register, dword indexed
  outportl(PCI_CAR32,pci_car32);       // restore CAR32
  enable();			       // enable interrupts
  return(data);
}

//***************** probe PCI devices and find all nVidia graphics cards ****
static int probe_devices()
{
  uint32_t did_vid;
  uint16_t vendor_id, device_id;
  int bus, dev, func, i=0, r;
  PCI_CONFIG_HEADER pCfg;
  for (bus = 0; bus < PCI_MAX_BUSES; bus++) // through all PCI buses
    for (dev = 0; dev < PCI_MAX_DEVICES; dev++) // through all PCI devices
      for (func = 0; func < PCI_MAX_FUNCS; func++) // through all PCI function blocks
	if ((did_vid = pci_read32(bus,dev,func,0)) != 0xFFFFFFFF) // register not empty
	  {
	  vendor_id = did_vid & 0xFFFF;
	  device_id = did_vid >> 16;
	  for (r = 0; r < 8; r++)      // read eight 32-bit registers
	    ((uint32_t *)&pCfg)[r] = pci_read32(bus,dev,func,r); // to PCI device header structure
	  // Check whether the vendor is nvidia and the BaseClass == VGA
	  if ((vendor_id == 0x10DE) && (pCfg.BaseClass == 0x03))
	    {
	    nvclock.card[i].device_id = device_id;
	    nvclock.card[i].arch = get_gpu_arch(nvclock.card[i].device_id);
	    nvclock.card[i].number = i;
	    nvclock.card[i].card_name = (char*)get_card_name(nvclock.card[i].device_id, &nvclock.card[i].gpu);
	    nvclock.card[i].reg_address = pCfg.BaseAddresses[0];
	    nvclock.card[i].devbusfn = func;
	    nvclock.card[i].irq = pCfg.InterruptLine;
	    nvclock.card[i].state = 0;
	    printf("Found VendorID: 0x%04X, DeviceID: 0x%04X @ PCI:%u:%u:%u, MMIO base: 0x%08X\r\n", vendor_id, device_id, bus, dev, func, nvclock.card[i].reg_address);
	    i++;
	    }
	  }
  nvclock.num_cards = i;	       // number of detected nVidia graphics cards
  if (i == 0)			       // if any nVidia adapter found
    nvclock.nv_errno |= NV_ERR_NO_DEVICES_FOUND; // set error flag
  return(i);
}

//***************** call DPMI physical memory mapping and allocate LDT descriptor for the memory block
void *map_physical_memory(uint32_t phys_base, uint32_t size) // phys_base must be page aligned
{							     // size will be aligned automatically
  void *p;
  if (size == 0)		       // if zero block size, do nothing
    return(NULL);
  if ((size & 0x0FFF) != 0)	       // if size is not page-aligned
    size = (size + 4096) & (~0x0FFF);  // make it page-aligned
  if ((p = valloc(size)) == NULL)      // allocate page-aligned memory (__djgpp_map_physical_memory required this)
    {				       // allocation failed
    printf("ERROR: cannot allocate %lu bytes for memory mapped device\n",size);
    exit(-1);			       // exit program
    }
  if (__djgpp_map_physical_memory(p,size,phys_base))
    {				       // failed to map physmem
    printf("ERROR: failed to map physical memory range: 0x%08lX - 0x%08lX\nDPMI 1.0 server or support of DPMI function 0508h is needed.\nIt doesn't work under Windows, use win32 version instead.\n",phys_base,phys_base+size-1);
    exit(-1);			       // exit program
    }
  return(p);			       // return pointer to memory block aliasing physical memory area
}

//***************** create mapping of nVidia MMIO registers to global nv_card structure
int map_mem(const char *dev_name)
{
  /* Map the registers of the nVidia chip */
  nv_card->PEXTDEV  = map_physical_memory(nv_card->reg_address + 0x101000, 0x1000);
  nv_card->PFB	    = map_physical_memory(nv_card->reg_address + 0x100000, 0x1000);
  /* normally pmc is till 0x2000 but extended it for nv40 */
  nv_card->PMC	    = map_physical_memory(nv_card->reg_address + 0x000000, 0x30000);
  nv_card->PCIO     = map_physical_memory(nv_card->reg_address + 0x601000, 0x2000);
  nv_card->PDISPLAY = map_physical_memory(nv_card->reg_address + NV_PDISPLAY_OFFSET, NV_PDISPLAY_SIZE);
  nv_card->PRAMDAC  = map_physical_memory(nv_card->reg_address + 0x680000, 0x4000); // 0x2000 for head 0, next 0x2000 for head 1
  nv_card->PRAMIN   = map_physical_memory(nv_card->reg_address + NV_PRAMIN_OFFSET, NV_PRAMIN_SIZE);
  nv_card->PROM     = map_physical_memory(nv_card->reg_address + 0x300000, 0x10000);
  /* On Geforce 8xxx cards it appears that the pci config header has been moved */
  if(nv_card->arch & NV5X)
    nv_card->PBUS   = map_physical_memory(nv_card->reg_address + 0x088000, 0x100);
  else
    nv_card->PBUS   = nv_card->PMC + 0x1800/4;
  nv_card->mem_mapped = 1;
  return(1);
}

//***************** free memory used for nVidia MMIO registers mapping ******
void unmap_mem()
{                                      // unmap is done automatically by DPMI server at exit
//  nv_card->mem_mapped = 0;	       // mapping disabled
}

int32_t pciReadLong(unsigned short devbusfn, long offset)
{
  return -1;
}

//***************** detect and initialize nVidia graphics cards *************
int init_nvclock()
{
  // Detect all nvidia cards; this needs to be done before creating directory and config file as that code needs card access
  if(!probe_devices())
    {
    // probe_devices takes care of the error as it isn't certain it failed because of there are no nvidia cards
    return(0);
    }
  if(!open_config())
    return(0);
  return(1);
}

#endif /* __DJGPP__ */
